/********************************************************************************
** Form generated from reading UI file 'editor.ui'
**
** Created by: Qt User Interface Compiler version 6.11.1
**
** WARNING! All changes made in this file will be lost when recompiling UI file!
********************************************************************************/

#ifndef UI_EDITOR_H
#define UI_EDITOR_H

#include <QtCore/QVariant>
#include <QtGui/QIcon>
#include <QtWidgets/QApplication>
#include <QtWidgets/QFrame>
#include <QtWidgets/QHBoxLayout>
#include <QtWidgets/QHeaderView>
#include <QtWidgets/QPushButton>
#include <QtWidgets/QSpacerItem>
#include <QtWidgets/QTableView>
#include <QtWidgets/QVBoxLayout>
#include <QtWidgets/QWidget>
#include <fcitxqti18nhelper.h>

QT_BEGIN_NAMESPACE

class Ui_Editor
{
public:
    QHBoxLayout *horizontalLayout;
    QTableView *macroTableView;
    QVBoxLayout *verticalLayout;
    QPushButton *addButton;
    QPushButton *deleteButton;
    QPushButton *clearButton;
    QFrame *line;
    QPushButton *importButton;
    QPushButton *exportButton;
    QSpacerItem *verticalSpacer;

    void setupUi(QWidget *Editor)
    {
        if (Editor->objectName().isEmpty())
            Editor->setObjectName("Editor");
        Editor->resize(375, 366);
        QIcon icon;
        QString iconThemeName = QString::fromUtf8("fcitx-unikey");
        if (QIcon::hasThemeIcon(iconThemeName)) {
            icon = QIcon::fromTheme(iconThemeName);
        } else {
            icon.addFile(QString::fromUtf8("."), QSize(), QIcon::Mode::Normal, QIcon::State::Off);
        }
        Editor->setWindowIcon(icon);
        horizontalLayout = new QHBoxLayout(Editor);
        horizontalLayout->setObjectName("horizontalLayout");
        macroTableView = new QTableView(Editor);
        macroTableView->setObjectName("macroTableView");
        macroTableView->setSelectionMode(QAbstractItemView::SingleSelection);
        macroTableView->setSelectionBehavior(QAbstractItemView::SelectRows);

        horizontalLayout->addWidget(macroTableView);

        verticalLayout = new QVBoxLayout();
        verticalLayout->setObjectName("verticalLayout");
        addButton = new QPushButton(Editor);
        addButton->setObjectName("addButton");
        QIcon icon1;
        iconThemeName = QString::fromUtf8("list-add");
        if (QIcon::hasThemeIcon(iconThemeName)) {
            icon1 = QIcon::fromTheme(iconThemeName);
        } else {
            icon1.addFile(QString::fromUtf8("."), QSize(), QIcon::Mode::Normal, QIcon::State::Off);
        }
        addButton->setIcon(icon1);

        verticalLayout->addWidget(addButton);

        deleteButton = new QPushButton(Editor);
        deleteButton->setObjectName("deleteButton");
        QIcon icon2;
        iconThemeName = QString::fromUtf8("list-remove");
        if (QIcon::hasThemeIcon(iconThemeName)) {
            icon2 = QIcon::fromTheme(iconThemeName);
        } else {
            icon2.addFile(QString::fromUtf8("."), QSize(), QIcon::Mode::Normal, QIcon::State::Off);
        }
        deleteButton->setIcon(icon2);

        verticalLayout->addWidget(deleteButton);

        clearButton = new QPushButton(Editor);
        clearButton->setObjectName("clearButton");
        QIcon icon3;
        iconThemeName = QString::fromUtf8("edit-delete");
        if (QIcon::hasThemeIcon(iconThemeName)) {
            icon3 = QIcon::fromTheme(iconThemeName);
        } else {
            icon3.addFile(QString::fromUtf8("."), QSize(), QIcon::Mode::Normal, QIcon::State::Off);
        }
        clearButton->setIcon(icon3);

        verticalLayout->addWidget(clearButton);

        line = new QFrame(Editor);
        line->setObjectName("line");
        line->setFrameShape(QFrame::Shape::HLine);
        line->setFrameShadow(QFrame::Shadow::Sunken);

        verticalLayout->addWidget(line);

        importButton = new QPushButton(Editor);
        importButton->setObjectName("importButton");
        QIcon icon4;
        iconThemeName = QString::fromUtf8("document-import");
        if (QIcon::hasThemeIcon(iconThemeName)) {
            icon4 = QIcon::fromTheme(iconThemeName);
        } else {
            icon4.addFile(QString::fromUtf8("."), QSize(), QIcon::Mode::Normal, QIcon::State::Off);
        }
        importButton->setIcon(icon4);

        verticalLayout->addWidget(importButton);

        exportButton = new QPushButton(Editor);
        exportButton->setObjectName("exportButton");
        QIcon icon5;
        iconThemeName = QString::fromUtf8("document-export");
        if (QIcon::hasThemeIcon(iconThemeName)) {
            icon5 = QIcon::fromTheme(iconThemeName);
        } else {
            icon5.addFile(QString::fromUtf8("."), QSize(), QIcon::Mode::Normal, QIcon::State::Off);
        }
        exportButton->setIcon(icon5);

        verticalLayout->addWidget(exportButton);

        verticalSpacer = new QSpacerItem(20, 40, QSizePolicy::Policy::Minimum, QSizePolicy::Policy::Expanding);

        verticalLayout->addItem(verticalSpacer);


        horizontalLayout->addLayout(verticalLayout);


        retranslateUi(Editor);

        QMetaObject::connectSlotsByName(Editor);
    } // setupUi

    void retranslateUi(QWidget *Editor)
    {
        Editor->setWindowTitle(fcitx::tr2fcitx("Unikey Macro Editor", nullptr));
        addButton->setText(fcitx::tr2fcitx("&Add", nullptr));
        deleteButton->setText(fcitx::tr2fcitx("&Delete", nullptr));
        clearButton->setText(fcitx::tr2fcitx("De&lete All", nullptr));
        importButton->setText(fcitx::tr2fcitx("&Import", nullptr));
        exportButton->setText(fcitx::tr2fcitx("&Export", nullptr));
    } // retranslateUi

};

namespace Ui {
    class Editor: public Ui_Editor {};
} // namespace Ui

QT_END_NAMESPACE

#endif // UI_EDITOR_H
