/********************************************************************************
** Form generated from reading UI file 'editor.ui'
**
** Created by: Qt User Interface Compiler version 6.11.1
**
** WARNING! All changes made in this file will be lost when recompiling UI file!
********************************************************************************/

#ifndef UI_EDITOR_H
#define UI_EDITOR_H

#include <QtCore/QVariant>
#include <QtGui/QIcon>
#include <QtWidgets/QApplication>
#include <QtWidgets/QComboBox>
#include <QtWidgets/QFormLayout>
#include <QtWidgets/QFrame>
#include <QtWidgets/QGroupBox>
#include <QtWidgets/QHBoxLayout>
#include <QtWidgets/QHeaderView>
#include <QtWidgets/QLabel>
#include <QtWidgets/QPushButton>
#include <QtWidgets/QSpacerItem>
#include <QtWidgets/QTableView>
#include <QtWidgets/QVBoxLayout>
#include <QtWidgets/QWidget>
#include <fcitxqti18nhelper.h>
#include "fcitxqtkeysequencewidget.h"

QT_BEGIN_NAMESPACE

class Ui_Editor
{
public:
    QVBoxLayout *verticalLayout_3;
    QGroupBox *groupBox;
    QHBoxLayout *horizontalLayout_2;
    QLabel *label;
    QComboBox *inputMethodBox;
    QPushButton *loadButton;
    QHBoxLayout *horizontalLayout;
    QVBoxLayout *verticalLayout_2;
    QGroupBox *groupBox_2;
    QFormLayout *formLayout;
    QLabel *label_3;
    QComboBox *categoryBox;
    QLabel *label_4;
    QComboBox *actionBox;
    QLabel *label_2;
    fcitx::FcitxQtKeySequenceWidget *keySequenceEdit;
    QTableView *keymapView;
    QVBoxLayout *verticalLayout;
    QPushButton *addButton;
    QPushButton *moveUpButton;
    QPushButton *moveDownButton;
    QPushButton *deleteButton;
    QPushButton *clearButton;
    QFrame *line;
    QPushButton *importButton;
    QPushButton *exportButton;
    QSpacerItem *verticalSpacer;

    void setupUi(QWidget *Editor)
    {
        if (Editor->objectName().isEmpty())
            Editor->setObjectName("Editor");
        Editor->resize(645, 555);
        QIcon icon;
        QString iconThemeName = QString::fromUtf8("fcitx-unikey");
        if (QIcon::hasThemeIcon(iconThemeName)) {
            icon = QIcon::fromTheme(iconThemeName);
        } else {
            icon.addFile(QString::fromUtf8("."), QSize(), QIcon::Mode::Normal, QIcon::State::Off);
        }
        Editor->setWindowIcon(icon);
        verticalLayout_3 = new QVBoxLayout(Editor);
        verticalLayout_3->setObjectName("verticalLayout_3");
        groupBox = new QGroupBox(Editor);
        groupBox->setObjectName("groupBox");
        horizontalLayout_2 = new QHBoxLayout(groupBox);
        horizontalLayout_2->setObjectName("horizontalLayout_2");
        label = new QLabel(groupBox);
        label->setObjectName("label");

        horizontalLayout_2->addWidget(label);

        inputMethodBox = new QComboBox(groupBox);
        inputMethodBox->setObjectName("inputMethodBox");
        QSizePolicy sizePolicy(QSizePolicy::Policy::MinimumExpanding, QSizePolicy::Policy::Fixed);
        sizePolicy.setHorizontalStretch(0);
        sizePolicy.setVerticalStretch(0);
        sizePolicy.setHeightForWidth(inputMethodBox->sizePolicy().hasHeightForWidth());
        inputMethodBox->setSizePolicy(sizePolicy);

        horizontalLayout_2->addWidget(inputMethodBox);

        loadButton = new QPushButton(groupBox);
        loadButton->setObjectName("loadButton");

        horizontalLayout_2->addWidget(loadButton);


        verticalLayout_3->addWidget(groupBox);

        horizontalLayout = new QHBoxLayout();
        horizontalLayout->setObjectName("horizontalLayout");
        verticalLayout_2 = new QVBoxLayout();
        verticalLayout_2->setObjectName("verticalLayout_2");
        groupBox_2 = new QGroupBox(Editor);
        groupBox_2->setObjectName("groupBox_2");
        formLayout = new QFormLayout(groupBox_2);
        formLayout->setObjectName("formLayout");
        label_3 = new QLabel(groupBox_2);
        label_3->setObjectName("label_3");

        formLayout->setWidget(0, QFormLayout::ItemRole::LabelRole, label_3);

        categoryBox = new QComboBox(groupBox_2);
        categoryBox->setObjectName("categoryBox");
        sizePolicy.setHeightForWidth(categoryBox->sizePolicy().hasHeightForWidth());
        categoryBox->setSizePolicy(sizePolicy);

        formLayout->setWidget(0, QFormLayout::ItemRole::FieldRole, categoryBox);

        label_4 = new QLabel(groupBox_2);
        label_4->setObjectName("label_4");

        formLayout->setWidget(1, QFormLayout::ItemRole::LabelRole, label_4);

        actionBox = new QComboBox(groupBox_2);
        actionBox->setObjectName("actionBox");
        sizePolicy.setHeightForWidth(actionBox->sizePolicy().hasHeightForWidth());
        actionBox->setSizePolicy(sizePolicy);

        formLayout->setWidget(1, QFormLayout::ItemRole::FieldRole, actionBox);

        label_2 = new QLabel(groupBox_2);
        label_2->setObjectName("label_2");

        formLayout->setWidget(2, QFormLayout::ItemRole::LabelRole, label_2);

        keySequenceEdit = new fcitx::FcitxQtKeySequenceWidget(groupBox_2);
        keySequenceEdit->setObjectName("keySequenceEdit");

        formLayout->setWidget(2, QFormLayout::ItemRole::FieldRole, keySequenceEdit);


        verticalLayout_2->addWidget(groupBox_2);

        keymapView = new QTableView(Editor);
        keymapView->setObjectName("keymapView");
        keymapView->setSelectionMode(QAbstractItemView::SingleSelection);
        keymapView->setSelectionBehavior(QAbstractItemView::SelectRows);

        verticalLayout_2->addWidget(keymapView);


        horizontalLayout->addLayout(verticalLayout_2);

        verticalLayout = new QVBoxLayout();
        verticalLayout->setObjectName("verticalLayout");
        addButton = new QPushButton(Editor);
        addButton->setObjectName("addButton");
        QIcon icon1;
        iconThemeName = QString::fromUtf8("document-edit");
        if (QIcon::hasThemeIcon(iconThemeName)) {
            icon1 = QIcon::fromTheme(iconThemeName);
        } else {
            icon1.addFile(QString::fromUtf8("."), QSize(), QIcon::Mode::Normal, QIcon::State::Off);
        }
        addButton->setIcon(icon1);

        verticalLayout->addWidget(addButton);

        moveUpButton = new QPushButton(Editor);
        moveUpButton->setObjectName("moveUpButton");
        QIcon icon2;
        iconThemeName = QString::fromUtf8("go-up");
        if (QIcon::hasThemeIcon(iconThemeName)) {
            icon2 = QIcon::fromTheme(iconThemeName);
        } else {
            icon2.addFile(QString::fromUtf8("."), QSize(), QIcon::Mode::Normal, QIcon::State::Off);
        }
        moveUpButton->setIcon(icon2);

        verticalLayout->addWidget(moveUpButton);

        moveDownButton = new QPushButton(Editor);
        moveDownButton->setObjectName("moveDownButton");
        QIcon icon3;
        iconThemeName = QString::fromUtf8("go-down");
        if (QIcon::hasThemeIcon(iconThemeName)) {
            icon3 = QIcon::fromTheme(iconThemeName);
        } else {
            icon3.addFile(QString::fromUtf8("."), QSize(), QIcon::Mode::Normal, QIcon::State::Off);
        }
        moveDownButton->setIcon(icon3);

        verticalLayout->addWidget(moveDownButton);

        deleteButton = new QPushButton(Editor);
        deleteButton->setObjectName("deleteButton");
        QIcon icon4;
        iconThemeName = QString::fromUtf8("list-remove");
        if (QIcon::hasThemeIcon(iconThemeName)) {
            icon4 = QIcon::fromTheme(iconThemeName);
        } else {
            icon4.addFile(QString::fromUtf8("."), QSize(), QIcon::Mode::Normal, QIcon::State::Off);
        }
        deleteButton->setIcon(icon4);

        verticalLayout->addWidget(deleteButton);

        clearButton = new QPushButton(Editor);
        clearButton->setObjectName("clearButton");
        QIcon icon5;
        iconThemeName = QString::fromUtf8("edit-delete");
        if (QIcon::hasThemeIcon(iconThemeName)) {
            icon5 = QIcon::fromTheme(iconThemeName);
        } else {
            icon5.addFile(QString::fromUtf8("."), QSize(), QIcon::Mode::Normal, QIcon::State::Off);
        }
        clearButton->setIcon(icon5);

        verticalLayout->addWidget(clearButton);

        line = new QFrame(Editor);
        line->setObjectName("line");
        line->setFrameShape(QFrame::Shape::HLine);
        line->setFrameShadow(QFrame::Shadow::Sunken);

        verticalLayout->addWidget(line);

        importButton = new QPushButton(Editor);
        importButton->setObjectName("importButton");
        QIcon icon6;
        iconThemeName = QString::fromUtf8("document-import");
        if (QIcon::hasThemeIcon(iconThemeName)) {
            icon6 = QIcon::fromTheme(iconThemeName);
        } else {
            icon6.addFile(QString::fromUtf8("."), QSize(), QIcon::Mode::Normal, QIcon::State::Off);
        }
        importButton->setIcon(icon6);

        verticalLayout->addWidget(importButton);

        exportButton = new QPushButton(Editor);
        exportButton->setObjectName("exportButton");
        QIcon icon7;
        iconThemeName = QString::fromUtf8("document-export");
        if (QIcon::hasThemeIcon(iconThemeName)) {
            icon7 = QIcon::fromTheme(iconThemeName);
        } else {
            icon7.addFile(QString::fromUtf8("."), QSize(), QIcon::Mode::Normal, QIcon::State::Off);
        }
        exportButton->setIcon(icon7);

        verticalLayout->addWidget(exportButton);

        verticalSpacer = new QSpacerItem(20, 40, QSizePolicy::Policy::Minimum, QSizePolicy::Policy::Expanding);

        verticalLayout->addItem(verticalSpacer);


        horizontalLayout->addLayout(verticalLayout);


        verticalLayout_3->addLayout(horizontalLayout);


        retranslateUi(Editor);

        QMetaObject::connectSlotsByName(Editor);
    } // setupUi

    void retranslateUi(QWidget *Editor)
    {
        Editor->setWindowTitle(fcitx::tr2fcitx("Unikey Keymap Editor", nullptr));
        groupBox->setTitle(fcitx::tr2fcitx("Built-in Input Methods", nullptr));
        label->setText(fcitx::tr2fcitx("Input Method:", nullptr));
        loadButton->setText(fcitx::tr2fcitx("&Load", nullptr));
        groupBox_2->setTitle(fcitx::tr2fcitx("Key definition", nullptr));
        label_3->setText(fcitx::tr2fcitx("Category:", nullptr));
        label_4->setText(fcitx::tr2fcitx("Action:", nullptr));
        label_2->setText(fcitx::tr2fcitx("Key", nullptr));
        addButton->setText(fcitx::tr2fcitx("&Add / Update", nullptr));
        moveUpButton->setText(fcitx::tr2fcitx("Move &Up", nullptr));
        moveDownButton->setText(fcitx::tr2fcitx("Move &Down", nullptr));
        deleteButton->setText(fcitx::tr2fcitx("&Delete", nullptr));
        clearButton->setText(fcitx::tr2fcitx("De&lete All", nullptr));
        importButton->setText(fcitx::tr2fcitx("&Import", nullptr));
        exportButton->setText(fcitx::tr2fcitx("&Export", nullptr));
    } // retranslateUi

};

namespace Ui {
    class Editor: public Ui_Editor {};
} // namespace Ui

QT_END_NAMESPACE

#endif // UI_EDITOR_H
